import inspect
from restfy.http import Request


class Handler:
    def __init__(self, func):
        self.func = func
        args = inspect.getfullargspec(func).annotations
        self.return_type = args.pop('return')
        self.parameters = args

    async def execute(self, properties):
        args = {}
        for key, kind in self.parameters.items():
            value = properties.get(key)
            if not value:
                raise Exception(f'Parameter {key}')
            if kind in [int, float, bool]:
                try:
                    value = kind(value)
                except Exception as e:
                    raise Exception(f'Error try cast value "{value}" {key} {kind}: {e}')
            args[key] = value
        ret = await self.func(**args)
        return ret


class Route:
    def __init__(self, name='', node='', path=None, handle=None, method='', prepare_data=True):
        self.properties = {}
        self.handlers = {}
        self.routes = {}
        self.variable = None
        self.is_variable = False
        self.variable_type = str
        self.name = name
        self.prepare_data: bool = prepare_data

    def add_node(self, path, handle, method='GET'):
        node = path.pop(0)
        if node.startswith('{'):
            if self.variable:
                route = self.variable
            else:
                route = Route()
                route.is_variable = True
                route.name = node[1:-1]
                self.variable = route
        else:
            route = self.routes.get(node, Route())
            route.name = node
            self.routes[node] = route
        if path:
            route.add_node(path=path, handle=handle, method=method)
        else:
            route.add_handler(Handler(handle), method)

    def add_handler(self, handle, method):
        self.handlers[method] = handle

    async def exec(self, request: Request):
        handler = self.handlers[request.method]
        if self.prepare_data and request.app.prepare_request_data:
            request.prepare_data()
        properties = {'request': request, **self.properties}
        return await handler.execute(properties)


class Router(Route):
    def __init__(self, base_url=''):
        super().__init__()
        self.base_url = base_url

    def add_route(self, path, handle, method='GET'):
        path = path[1:].split('/')
        if len(path) == 1 and path[0] == '':
            self.add_handler(Handler(handle), method)
        else:
            self.add_node(path=path, handle=handle, method=method)

    def register_router(self, path, router):
        nodes = path[1:].split('/')
        if len(nodes) == 1 and nodes[0] == '':
            self.routes = router.routes
            self.variable = router.variable
            self.is_variable = router.is_variable
        else:
            routes = self.routes
            while True:
                node = nodes.pop(0)
                if len(nodes) == 0:
                    routes[node] = router
                    break
                else:
                    if routes[node].routes:
                        routes = routes[node].routes
                    else:
                        routes = routes[node].variable

    def match(self, url, method):
        nodes = url[1:].split('/')
        if len(nodes) == 1 and nodes[0] == '':
            return self
        routes = self.routes
        variable = self.variable
        properties = {}
        while len(nodes) > 0:
            node = nodes.pop(0)
            route = routes.get(node, None)
            if not route:
                if variable:
                    route = variable
                    properties[route.name] = node
                else:
                    break
            routes = route.routes
            variable = route.variable
        if route:
            if method in route.handlers:
                route.properties = properties
            else:
                route = None
        return route

    def get(self, path):
        def wrapper(func):
            self.add_route(path, handle=func)
            return func
        return wrapper

    def post(self, path):
        def wrapper(func):
            self.add_route(path, handle=func, method='POST')
            return func
        return wrapper

    def put(self, path):
        def wrapper(func):
            self.add_route(path, handle=func, method='PUT')
            return func
        return wrapper

    def delete(self, path):
        def wrapper(func):
            self.add_route(path, handle=func, method='DELETE')
            return func
        return wrapper

    def patch(self, path):
        def wrapper(func):
            self.add_route(path, handle=func, method='PATCH')
            return func
        return wrapper

    def options(self, path):
        def wrapper(func):
            self.add_route(path, handle=func, method='OPTIONS')
            return func
        return wrapper

    def head(self, path):
        def wrapper(func):
            self.add_route(path, handle=func, method='HEAD')
            return func
        return wrapper





class AccessControl:
    allow_origin = '*'
    expose_headers = []
    max_age = 0.0
    allow_credentials = False
    allow_methods = []
    allow_headers = []

    def get_response_headers(self):
        headers = {
            'Access-Control-Allow-Origin': self.allow_origin,
        }
        if self.allow_methods:
            headers['Access-Control-Allow-Methods'] = ','.join(self.allow_methods)
        if self.allow_headers:
            headers['Access-Control-Allow-Headers'] = ','.join(self.allow_headers)
        if self.allow_credentials:
            headers['Access-Control-Allow-Credentials'] = self.allow_credentials
        if self.max_age:
            headers['Access-Control-Max-Age'] = self.max_age
        if self.expose_headers:
            headers['Access-Control-Expose-Headers'] = self.expose_headers
        return headers





import json

status_title = {
    200: 'OK',
    201: 'CREATED',
    202: 'ACCEPTED',
    203: 'NON-AUTHORITATIVE INFORMATION',
    204: 'NO CONTENT',
    205: 'RESET CONTENT',
    206: 'PARTIAL CONTENT',
    400: 'BAD REQUEST',
    401: 'UNAUTHORIZED',
    402: 'PAYMENT REQUIRED',
    403: 'FORBIDDEN',
    404: 'NOT FOUND',
    405: 'METHOD NOT ALLOWED',
    406: 'NOT ACCEPTABLE',
    407: 'PROXY AUTHENTICATION REQUIRED',
    408: 'REQUEST TIMEOUT',
    409: 'CONFLIT',
    410: 'GONE',
    500: 'INTERNAL SERVER ERROR',
    501: 'NOT IMPLEMENTED',
    502: 'BAD GATEWAY',
    503: 'SERVICE UNAVAILABLE',
    504: 'GATEWAY TIMEOUT',
    505: 'HTTP VERSION NOT SUPORTED'
}


class Response:
    def __init__(
            self,
            data='',
            *,
            status: int = 200,
            content_type: str = '',
            headers: dict = None
    ):
        self.version = 'HTTP/1.1'
        self.status = status
        self.data = data if status != 204 else None
        self.headers = {}
        self.content_type = content_type
        self._prepare_headers(headers)

    def render(self):
        title = status_title.get(self.status, 'STATUS WITHOUT TITLE')
        headers = '\r\n'.join([f"{k}:{v}" for k, v in self.headers.items()])
        body = self.data
        content = f'{self.version} {self.status} {title}\r\n{headers}\r\n\r\n{body}'
        return content.encode()

    def _prepare_headers(self, headers):
        if not headers:
            headers = {}
        if isinstance(self.data, dict) or isinstance(self.data, list):
            self.data = json.dumps(self.data)
            self.headers['Content-Type'] = 'application/json'
        elif isinstance(self.data, str):
            self.headers['Content-Type'] = 'text/plain'
        if self.data:
            self.headers['Content-Length'] = len(self.data)
        if self.content_type:
            self.headers['Content-Type'] = self.content_type
        self.headers.update(headers)



from .http import Request, Response


class Middleware:
    def __init__(self):
        self.next = None

    async def exec(self, request) -> Response:
        response = await self.forward(request)
        return response

    async def forward(self, request) -> Response:
        response = await self.next.exec(request)
        return response


class File:
    def __init__(self, name: str, kind: str, content=None, base64=False):
        self.content = content
        self.name = name
        self.type = kind
        self.base64 = base64

    def save(self, path):
        with open(path, 'wb') as f:
            f.write(self.content)

    def open(self, path):
        with open(path, 'rb') as f:
            self.content = f.read()


import json
from restfy.file import File


class Request:
    def __init__(self, method, version):
        self.app = None
        self.method = method
        self.url = ''
        self.version = version
        self.body = None
        self.type = ''
        self.query = ''
        self.length = 0
        self.headers = {}
        self.files = {}
        self.origin = ''
        self.request_method = ''
        self.request_headers = ''
        self.preflight = False
        self.multipart = False
        self.boundary = ''
        self.data = {}

    def add_header(self, key, value):
        self.headers[key] = value
        if key == 'Content-Type':
            if 'multipart/form-data' in value:
                self.multipart = True
                (content, boundary) = value.split(';')
                self.type = content.strip()
                self.boundary = boundary.replace('boundary=', '').strip()
            else:
                self.type = value
        elif key == 'Content-Length':
            self.length = int(value)
        elif key == 'Origin':
            self.origin = value
            self.preflight = True if self.method == 'OPTIONS' else False
        elif key == 'Access-Control-Request-Method':
            self.request_method = value
        elif key == 'Access-Control-Request-Headers':
            self.request_headers = value

    def dict(self):
        if self.body:
            if self.type == 'application/json':
                return json.loads(self.body)
            elif self.type == 'multipart/form-data':
                return self._process_form_data()
            elif self.type == 'application/x-www-form-urlencoded':
                return self._url_decoded_data()
        return {}

    def args(self):
        args = {}
        if self.query:
            pairs = self.query.split('&')
            for pair in pairs:
                (key, value) = pair.split('=')
                args[key] = value
        return args

    def prepare_url(self, url):
        if '?' in url:
            (path, query) = url.split('?')
        else:
            path = url
            query = ''
        self.url = path
        self.query = query

    def prepare_data(self):
        self.data = self.dict()

    def _process_form_data(self):
        data = {}
        parts = self.body.split(f'--{self.boundary}'.encode())
        for part in parts:
            if not part or part == b'--\r\n':
                continue
            if b'filename=' in part:
                splt = part.split(b';', maxsplit=2)
                key = splt[1].decode().strip()[6:-1]
                (info, content) = splt[2].split(b'\r\n\r\n', maxsplit=1)
                (filename, kind) = info.decode().split('\r\n')
                filename = filename.strip()[10:-1]
                kind = kind.strip()[14:]
                file = File(name=filename, kind=kind, content=content)
                self.files[key] = file
            else:
                splt = part.split(b';')
                key, value = splt[1].split(b'\r\n\r\n')
                key = key.decode().replace('name=', '').strip()[1:-1]
                data[key] = value.strip().decode()
        return data

    def _url_decoded_data(self):
        data = {}
        pairs = self.body.decode().split('&')
        for pair in pairs:
            (key, value) = pair.split('=')
            data[key] = value
        return data



import asyncio
import datetime
import time
import inspect
from typing import List

from .http import Request, Response, AccessControl
from .router import Router, Route
from .middleware import Middleware


class Application:
    def __init__(
            self,
            base_url: str = '',
            prepare_request_data: bool = True
    ):
        self.router = Router(base_url=base_url)
        self.cors = AccessControl()
        self.middlewares: List[Middleware] = []
        self.prepare_request_data = prepare_request_data

    def add_route(self, path, handle, method='GET'):
        self.router.add_route(path, handle, method)

    def register_router(self, path, router):
        self.router.register_router(path, router)

    async def handler(self, reader: asyncio.streams.StreamReader, writer: asyncio.streams.StreamWriter):
        start = datetime.datetime.now()
        ini = time.time()
        data = await reader.readline()
        (method, url, version) = data.decode().replace('\n', '').split(' ')
        request = Request(method=method, version=version)
        request.prepare_url(url)
        try:
            while True:
                line = await reader.readline()
                header = line.decode()
                if header == '\r\n':
                    break
                header = header.replace('\r\n', '')
                splt = header.split(':', maxsplit=1)
                request.add_header(key=splt[0].strip(), value=splt[1].strip())
            if request.length:
                size = request.length
                content = b''
                while True:
                    content += await reader.read(size)
                    size = request.length - len(content)
                    if size == 0:
                        break
                request.body = content
            if request.preflight:
                response = Response(status=204)
                response.headers.update(self.cors.get_response_headers())
            else:
                if route := self.router.match(request.url, method):
                    request.app = self
                    response = await self.execute_middlewares(route, request)
                    if request.origin:
                        response.headers.update(self.cors.get_response_headers())
                else:
                    response = Response(status=404)
        except Exception as e:
            response = Response({'message': 'Internal server error', 'detail': str(e)}, status=500)
        writer.write(response.render())
        await writer.drain()
        writer.close()
        diff = time.time() - ini
        print(f'{[start.isoformat()]} {method} {url} --> {response.status}: {diff * 1000} ms')

    async def execute_middlewares(self, route: Route, request: Request) -> Response:
        if self.middlewares:
            self.middlewares[-1].next = route
            response = await self.middlewares[0].exec(request)
        else:
            response = await route.exec(request)
        return response

    def register_middleware(self, middleware):
        instance = middleware()
        if self.middlewares:
            self.middlewares[-1].next = instance
        self.middlewares.append(instance)

    def get(self, path):
        return self.router.get(path)

    def post(self, path):
        return self.router.post(path)

    def put(self, path):
        return self.router.put(path)

    def delete(self, path):
        return self.router.delete(path)

    def patch(self, path):
        return self.router.patch(path)

    def options(self, path):
        return self.router.options(path)

    def head(self, path):
        return self.router.head(path)




from .application import Application
from .router import Router
from .server import Server
from .http import Request, Response
from .middleware import Middleware


__all__ = ('Application', 'Server', 'Router', 'Middleware')



import asyncio


class Server:
    def __init__(self, app=None, host='0.0.0.0', port=7777):
        self.app = app
        self.host = host
        self.port = port

    async def serve(self):
        print(f'RESTFY ON {self.port}')
        server = await asyncio.start_server(self.app.handler, self.host, self.port)
        async with server:
            await server.serve_forever()

    def run(self):
        asyncio.run(self.serve())
